import struct


def format_sid(sid: bytes) -> str:
    parts = [struct.unpack('I', sid[1:5])[0],
             struct.unpack('I', sid[8:12])[0],
             struct.unpack('I', sid[12:16])[0],
             struct.unpack('I', sid[16:20])[0],
             struct.unpack('I', sid[20:24])[0],
             struct.unpack('I', sid[24:28])[0]
             ]

    return 'S-%d-%s' % (sid[0], '-'.join(map(str, parts)))


def parse_sid(sid: str) -> bytes:
    parts = b'\x01' + \
            struct.pack('I', int(sid[4])) + \
            b'\x00\x00\x05'

    sections = sid.split('-')
    for section in sections[3:]:
        parts += struct.pack('I', int(section))

    return parts



"""
 * Convert a microsoft timestamp to UNIX timestamp
 * http://www.morecavalier.com/index.php?whom=Apps%2FLDAP+timestamp+converter
 * @param int ad_date Microsoft timestamp
 * @return int UNIX timestamp
 """
from datetime import datetime


def microsoft_timestamp_to_unix(ad_date):
    ad_date = int(ad_date)
    if not ad_date:
        raise ValueError('Invalid date')

    secsAfterADEpoch = ad_date / 10000000
    AD2Unix = ((1970 - 1601) * 365 - 3 + round((1970 - 1601) / 4)) * 86400

    """Why -3 ?
    If the year is the last year of a century, eg. 1700, 1800, 1900, 2000,
    then it is only a leap year if it is exactly divisible by 400.
    Therefore, 1900 wasn't a leap year but 2000 was."""

    unixTimeStamp = int(secsAfterADEpoch - AD2Unix)

    return unixTimeStamp


def parse_date(ad_timestamp) -> datetime:
    timestamp = microsoft_timestamp_to_unix(ad_timestamp)
    return datetime.fromtimestamp(timestamp)


def find_flags(flag):
    flag_list = {
        1: 'SCRIPT',
        2: 'ACCOUNTDISABLE',
        8: 'HOMEDIR_REQUIRED',
        16: 'LOCKOUT',
        32: 'PASSWD_NOTREQD',
        64: 'PASSWD_CANT_CHANGE',
        128: 'ENCRYPTED_TEXT_PWD_ALLOWED',
        256: 'TEMP_DUPLICATE_ACCOUNT',
        512: 'NORMAL_ACCOUNT',
        2048: 'INTERDOMAIN_TRUST_ACCOUNT',
        4096: 'WORKSTATION_TRUST_ACCOUNT',
        8192: 'SERVER_TRUST_ACCOUNT',
        65536: 'DONT_EXPIRE_PASSWORD',
        131072: 'MNS_LOGON_ACCOUNT',
        262144: 'SMARTCARD_REQUIRED',
        524288: 'TRUSTED_FOR_DELEGATION',
        1048576: 'NOT_DELEGATED',
        2097152: 'USE_DES_KEY_ONLY',
        4194304: 'DONT_REQ_PREAUTH',
        8388608: 'PASSWORD_EXPIRED',
        16777216: 'TRUSTED_TO_AUTH_FOR_DELEGATION',
        67108864: 'PARTIAL_SECRETS_ACCOUNT'
    }
    flags = []
    for i in range(0, 27):

        if flag & (1 << i):
            flags.append(1 << i)

    flags_output = {}
    for v in flags:
        flags_output[v] = flag_list[v]

    return flags_output



# Generated by Django 3.1.1 on 2021-01-25 18:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ad_import', '0007_name_fields'),
    ]

    operations = [
        migrations.RenameField(
            model_name='server',
            old_name='objectSid',
            new_name='objectSidString',
        ),
        migrations.RenameField(
            model_name='workstation',
            old_name='objectSid',
            new_name='objectSidString',
        ),
        migrations.RenameField(
            model_name='user',
            old_name='objectSid',
            new_name='objectSidString',
        ),
        migrations.AddField(
            model_name='user',
            name='objectSid',
            field=models.BinaryField(),
        ),
        migrations.AddField(
            model_name='server',
            name='objectSid',
            field=models.BinaryField(),
        ),
        migrations.AddField(
            model_name='workstation',
            name='objectSid',
            field=models.BinaryField(),
        ),
    ]



from ldap.controls import SimplePagedResultsControl


# https://stackoverflow.com/a/61829690/2630074
def paginated_query(connect, attributes, base, scope, search_flt, result_class=None):
    # how many users to search for in each page,
    # this depends on the server maximum setting
    # (default highest value is 1000)
    page_size = 500

    req_ctrl = SimplePagedResultsControl(
        criticality=True, size=page_size, cookie=""
    )
    msgid = connect.search_ext(
        base=base,
        scope=scope,
        filterstr=search_flt,
        attrlist=attributes,
        serverctrls=[req_ctrl],
    )

    total_results = []
    pages = 0
    while True:  # loop over all of the pages using the same cookie,
        # otherwise the search will fail
        pages += 1
        rtype, rdata, rmsgid, serverctrls = connect.result3(msgid)
        for user in rdata:
            if not result_class:
                total_results.append(user)
            else:
                total_results.append(result_class(user))

        pctrls = [
            c
            for c in serverctrls
            if c.controlType == SimplePagedResultsControl.controlType
        ]
        if pctrls:
            if pctrls[
                0
            ].cookie:  # Copy cookie from response control to request control
                req_ctrl.cookie = pctrls[0].cookie
                msgid = connect.search_ext(
                    base=base,
                    scope=scope,
                    filterstr=search_flt,
                    attrlist=attributes,
                    serverctrls=[req_ctrl],
                )
            else:
                break
        else:
            break
    return total_results



from django.db import models

from ad_import.models import Directory

IMPORT_TARGETS = [('ad_import.models.workstation', 'Workstation'),
                  ('ad_import.models.server', 'Server'),
                  ('ad_import.models.user', 'User'),
                  ('ad_import.models.group', 'Group')]


class Query(models.Model):
    name = models.CharField(max_length=50, blank=True, null=True)
    directory = models.ForeignKey(Directory, on_delete=models.PROTECT, related_name='queries')
    query = models.CharField(max_length=200)
    base_dn = models.CharField('Base DN', max_length=200, blank=True, null=True)
    type = models.CharField(max_length=15, choices=[('user', 'User'), ('computer', 'Computer'), ('group', 'Group')])
    target = models.CharField(max_length=200, choices=IMPORT_TARGETS)

    def __str__(self):
        return '%s %s' % (self.directory.dn, self.query)

    class Meta:
        verbose_name = 'LDAP query'
        verbose_name_plural = 'LDAP queries'



# Generated by Django 3.1.8 on 2021-08-13 17:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('ad_import', '0008_sid_bytes'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='user',
            options={'permissions': [('show_last_logon', 'Can show user last logon')]},
        ),
        migrations.AlterField(
            model_name='user',
            name='manager',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='manages',
                                    to='ad_import.user', verbose_name='Leder'),
        ),
    ]



from django.contrib.auth.decorators import permission_required
from django.shortcuts import render

from ad_import.models import User


@permission_required('ad_import.view_user', raise_exception=True)
def view_user(request):
    user_id = request.GET.get('id')
    user = User.objects.get(id=user_id)
    return render(request, 'ad_import/user.html', {'user': user})



from ad_import.load_data import LoadWorkstations
from ad_import.management.commands import ADBaseCommand


class Command(ADBaseCommand):
    def handle(self, *args, **options):
        self.load(options, LoadWorkstations)



# Generated by Django 3.1.1 on 2020-10-01 14:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ad_import', '0005_directory_related_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='manager',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ad_import.user', verbose_name='Leder'),
        ),
    ]



import unittest
import ad.sid


class TestSID(unittest.TestCase):
    byte_value = b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xa3<\x9c\x1e\x8a\x15iT\xa2\x18\x94\xda\xab\x07\x01\x00'
    string_value = 'S-1-5-21-513555619-1416172938-3667138722-67499'

    def test_format_sid(self):
        self.assertEqual(self.string_value, sid.format_sid(self.byte_value))

    def test_parse_sid(self):
        self.assertEqual(self.byte_value, sid.parse_sid(self.string_value))



# Generated by Django 3.1 on 2020-09-18 14:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ad_import', '0003_server_workstation'),
    ]

    operations = [
        migrations.AlterField(
            model_name='query',
            name='target',
            field=models.CharField(choices=[('ad_import.models.workstation', 'Workstation'), ('ad_import.models.server', 'Server'), ('ad_import.models.user', 'User'), ('ad_import.models.group', 'Group')], max_length=200),
        ),
    ]



from abc import ABC
from typing import Optional

from ad_import.ad import ADObject
from ad_import.load_data import LoadAd
from ad_import.models import Computer, Query


class LoadComputers(LoadAd, ABC):
    fields = ['cn',
              'description',
              'distinguishedName',
              'lastLogon',
              'logonCount',
              'name',
              'objectGUID',
              'objectSid',
              'operatingSystem',
              'operatingSystemHotfix',
              'operatingSystemServicePack',
              'operatingSystemVersion',
              'pwdLastSet',
              'userAccountControl',
              'whenChanged',
              'whenCreated',
              ]

    model = Computer

    def load_object(self, user_data: ADObject) -> Optional[Computer]:
        """
        Load a server AD object into the database
        """
        computer: Computer
        computer, exist = self.get_object(user_data)

        computer.cn = user_data.string('cn')
        computer.description = user_data.string('description')
        computer.distinguishedName = user_data.string('distinguishedName')
        computer.lastLogon = user_data.date('lastLogon')
        computer.logonCount = user_data.numeric('logonCount')
        computer.name = user_data.string('name')
        # computer.objectGUID = user_data.bytes('objectGUID')
        computer.objectGUID = user_data.bytes_int('objectGUID')
        computer.objectSid = user_data.bytes('objectSid')
        computer.objectSidString = user_data.bytes_int('objectSid')
        computer.operatingSystem = user_data.string('operatingSystem')
        computer.operatingSystemHotfix = user_data.string('operatingSystemHotfix')
        computer.operatingSystemServicePack = user_data.string('operatingSystemServicePack')
        computer.operatingSystemVersion = user_data.string('operatingSystemVersion')
        computer.pwdLastSet = user_data.date('pwdLastSet')
        computer.userAccountControl = user_data.numeric('userAccountControl')
        computer.whenChanged = user_data.date_string('whenChanged')
        computer.whenCreated = user_data.date_string('whenCreated')

        if not computer.userAccountControl:
            print('%s has no userAccountControl' % computer)
            return

        if computer.disabled() and not exist:
            print('%s is disabled' % computer)
            return
        else:
            computer.save()

        return computer

    def load(self, query: Query):
        entries = self.run_query(query)

        for entry in entries:
            self.load_object(entry)



# Generated by Django 3.1.1 on 2020-11-26 15:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ad_import', '0006_manager_set_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='directory',
            name='dns_name',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='Domain name'),
        ),
        migrations.AddField(
            model_name='directory',
            name='short_name',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='Domain name (pre-Windows 2000)'),
        ),
    ]



from pprint import pprint
from typing import Optional

from ad_import.models import User
from . import LoadAd


class LoadUsers(LoadAd):
    fields = ['accountExpires',
              'cn',
              'company',
              'department',
              'departmentNumber',
              'description',
              'displayName',
              'distinguishedName',
              'employeeID',
              'employeeNumber',
              'givenName',
              'homeDirectory',
              'homeDrive',
              'ipPhone',
              'lastLogon',
              'logonCount',
              'mail',
              'manager',
              'middleName',
              'mobile',
              'name',
              'objectGUID',
              'objectSid',
              'physicalDeliveryOfficeName',
              'pwdLastSet',
              'sAMAccountName',
              'scriptPath',
              'sn',
              'telephoneNumber',
              'title',
              'userAccountControl',
              'userPrincipalName',
              'whenChanged',
              'whenCreated',
              ]

    model = User

    def load_user(self, user_data) -> Optional[User]:
        user, exist = self.get_object(user_data)

        if user_data.field('userAccountControl') is None:
            pprint(user_data.object)
            return

        if user_data.numeric('userAccountControl') & 2 == 2:  # Check if account is disabled
            if not exist:
                print('%s is disabled' % user_data.string('displayName'))
                return

        # user.accountExpires = user_data.date('accountExpires')
        user.cn = user_data.string('cn')
        user.company = user_data.string('company')
        user.department = user_data.string('department')
        user.departmentNumber = user_data.string('departmentNumber')
        user.description = user_data.string('description')
        user.displayName = user_data.string('displayName')
        user.distinguishedName = user_data.string('distinguishedName')
        user.employeeID = user_data.string('employeeID')
        user.employeeNumber = user_data.string('employeeNumber')
        user.givenName = user_data.string('givenName')
        user.homeDirectory = user_data.string('homeDirectory')
        user.homeDrive = user_data.string('homeDrive')
        user.ipPhone = user_data.string('ipPhone')
        user.lastLogon = user_data.date('lastLogon')
        user.logonCount = user_data.numeric('logonCount')
        user.mail = user_data.string('mail')
        user.middleName = user_data.string('middleName')
        user.mobile = user_data.string('mobile')
        user.name = user_data.string('name')
        # user.objectGUID = user_data.bytes('objectGUID')
        user.objectGUID = user_data.bytes_int('objectGUID')
        user.objectSid = user_data.bytes('objectSid')
        user.objectSidString = user_data.bytes_int('objectSid')
        user.physicalDeliveryOfficeName = user_data.string('physicalDeliveryOfficeName')
        user.pwdLastSet = user_data.date('pwdLastSet')
        user.sAMAccountName = user_data.string('sAMAccountName')
        user.scriptPath = user_data.string('scriptPath')
        user.sn = user_data.string('sn')
        user.telephoneNumber = user_data.string('telephoneNumber')
        user.title = user_data.string('title')
        user.userAccountControl = user_data.numeric('userAccountControl')
        user.userPrincipalName = user_data.string('userPrincipalName')
        user.whenChanged = user_data.date_string('whenChanged')
        user.whenCreated = user_data.date_string('whenCreated')

        if user.disabled() and not exist:
            print('%s is disabled' % user)
            return
        else:
            manager = user_data.string('manager')
            if manager:
                try:
                    user.manager = User.objects.get(directory=self.directory, distinguishedName=manager)
                except User.DoesNotExist:
                    print('Manager %s does not exist' % manager)

            user.save()

        return user

    def load(self, query):
        entries = self.run_query(query)

        for user_data in entries:
            self.load_user(user_data)



from __future__ import annotations

from django.db import models
from django.db.models import Q


class Directory(models.Model):
    name = models.CharField(max_length=100, null=True, blank=True)
    dns_name = models.CharField('Domain name',
                                max_length=100, null=True, blank=True)
    short_name = models.CharField('Domain name (pre-Windows 2000)',
                                  max_length=100, null=True, blank=True)
    dn = models.CharField(max_length=100)
    dc = models.GenericIPAddressField()
    username = models.CharField(max_length=100)
    password = models.CharField(max_length=100)
    ldaps = models.BooleanField(default=False)

    def __str__(self):
        return self.dn

    class Meta:
        # verbose_name = 'Directory'
        verbose_name_plural = 'directories'

    @staticmethod
    def find_directory(name: str) -> Directory:
        return Directory.objects.get(
            Q(dns_name__iexact=name) |
            Q(name__iexact=name) |
            Q(short_name__iexact=name)
        )



# Generated by Django 3.1 on 2020-09-18 20:31

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ad_import', '0004_target_choices'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='directory',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='users', to='ad_import.directory'),
        ),
    ]



from abc import ABC
from typing import Type

from django.core.management.base import BaseCommand
from django.db import IntegrityError
from django.db.models import ProtectedError

from ad_import.load_data import LoadAd
from ad_import.models import Directory


class ADBaseCommand(BaseCommand, ABC):
    load_class: Type[LoadAd] = None

    def add_arguments(self, parser):
        directories = Directory.objects.all().values_list('name', flat=True)
        parser.add_argument('directory', nargs='+',
                            choices=list(directories), help='AD')

    @staticmethod
    def get_directories(options):
        if type(options['directory']) == list:
            directories = []
            for directory_string in options['directory']:
                directory = Directory.find_directory(directory_string)
                directories.append(directory)
            return directories
        elif type(options['directory']) == str:
            return Directory.find_directory(options['directory'])

    @staticmethod
    def run_queries(directory: Directory, load_class: Type[LoadAd]):
        print('Loading from %s' % directory)
        load = load_class()
        load.connect(directory=directory)
        for query in load.queries:
            print('Run query %s' % query)
            load.load(query)

        if hasattr(load, 'get_inactive'):
            inactive = load.get_inactive()
            try:
                inactive.delete()
            except ProtectedError:
                print('Protected relations')
            except IntegrityError:
                print('Integrity error')

    def load(self, options: dict, load_class: Type[LoadAd]):
        for directory in self.get_directories(options):
            self.run_queries(directory, load_class)
